@@ -4,18 +4,39 @@
  GnomeICU
 *********************************/
 
-#include "gnomeicu.h"
+#include "common.h"
 #include "filexfer.h"
-#include "util.h"
-#include "tcp.h"
-#include "packetprint.h"
-#include "history.h"
-#include "gtkfunc.h"
 #include "filexferdlg.h"
+#include "gnomeicu.h"
+#include "gtkfunc.h"
+#include "history.h"
+#include "packetprint.h"
+#include "tcp.h"
+#include "util.h"
+
+#include <fcntl.h>
+#include <netinet/in.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <unistd.h>
+#include <gtk/gtk.h>
+#include <libgnomeui/libgnomeui.h>
 
 static int ft_next_file( XferInfo *xfer );
 static int ft_handshake( GIOChannel *source, GIOCondition condition,
                          XferInfo *xfer );
+static int ft_filereadserver( GIOChannel *source, GIOCondition condition,
+                       XferInfo *xfer );
+static int ft_filereadclient( GIOChannel *source, GIOCondition condition,
+                       XferInfo *xfer );
+static void ft_sendfileinfo( XferInfo *xfer );
+static void ft_skip_file( GtkWidget *, XferInfo *xfer);
+static int ft_sendpartfile( GIOChannel *source, GIOCondition condition,
+                     XferInfo *xfer );
+
 
 int ft_connectfile( XferInfo *xfer )
 {
@@ -23,30 +44,19 @@ int ft_connectfile( XferInfo *xfer )
   DWORD localport;
   struct sockaddr_in local, remote;
   int sizeofSockaddr = sizeof( struct sockaddr );
-  BYTE *buffer;
-  int index, psize;
+/*   BYTE *buffer; */
+/*   int index, psize; */
 
-  struct {
-    BYTE id; /* 0xff */
-    BYTE version[4]; /* 0x00000003 */
-    BYTE port[4]; /* TCP listen port */
-    BYTE uin[4]; /* UIN */
-    BYTE ipa[4]; /* IP A */
-    BYTE ipb[4]; /* IP B */
-    BYTE four; /* 0x04 */
-    BYTE listenport[4]; /* File listen port */
-  } handshake;
-
-  struct {
-    BYTE psize[2];
-    BYTE id; /* 0x00 */
-    BYTE zero[4]; /* 0x00000000 */
-    BYTE n_files[4]; /* Number of files */
-    BYTE n_bytes[4]; /* Total number of bytes */
-    BYTE speed[4]; /* Speed of transfer, 0x00 - 0x64 */
-    BYTE nick_len[2]; /* Length of nick */
-    /* Nickname */
-  } xfer0;
+/*   struct { */
+/*     BYTE id; /\* 0xff *\/ */
+/*     BYTE version[4]; /\* 0x00000003 *\/ */
+/*     BYTE port[4]; /\* TCP listen port *\/ */
+/*     BYTE uin[4]; /\* UIN *\/ */
+/*     BYTE ipa[4]; /\* IP A *\/ */
+/*     BYTE ipb[4]; /\* IP B *\/ */
+/*     BYTE four; /\* 0x04 *\/ */
+/*     BYTE listenport[4]; /\* File listen port *\/ */
+/*   } handshake; */
 
 #ifdef TRACE_FUNCTION
   g_print( "ft_connectfile\n" );
@@ -57,15 +67,15 @@ int ft_connectfile( XferInfo *xfer )
   if( xfer->sock_fd > 0 )
     return xfer->sock_fd;
 
-  if (!xfer->remote_contact->current_ip ||
-            !xfer->remote_contact->has_direct_connect)
+  if (!xfer->remote_contact->current_ip )
+    /* || !xfer->remote_contact->has_direct_connect) */
     return -1;
 
   xfer->sock_fd = socket( PF_INET, SOCK_STREAM, 0 );
   if( xfer->sock_fd == -1 )
     return -1;
 
-  fcntl( xfer->sock_fd, O_NONBLOCK );
+  /* fcntl( xfer->sock_fd, O_NONBLOCK ); */
 
   local.sin_family = AF_INET;
   remote.sin_family = AF_INET;
@@ -85,24 +95,22 @@ int ft_connectfile( XferInfo *xfer )
   getsockname( xfer->sock_fd, (struct sockaddr*)&local, &sizeofSockaddr );
   localport = g_ntohs( local.sin_port );
 
+
+  TCPSendHandShake(xfer->remote_contact, xfer->remote_contact->direct_cookie,
+		   xfer->sock_fd);
+
+  /*
   handshake.id = 0xff;
   DW_2_Chars( handshake.version, 0x00000003 );
   DW_2_Chars( handshake.port, our_port );
-  DW_2_Chars( handshake.uin, our_info->uin );
+  DW_2_Chars( handshake.uin, atoi (our_info->uin) );
   DW_2_IP( handshake.ipa, our_ip );
   DW_2_IP( handshake.ipb, LOCALHOST );
   handshake.four = 0x04;
   DW_2_Chars( handshake.listenport, localport );
+  */
 
-  Word_2_Chars( xfer0.psize, sizeof( xfer0 ) - 2 + strlen( our_info->nick ) + 1 );
-  xfer0.id = 0x00;
-  DW_2_Chars( xfer0.zero, 0x00000000 );
-  DW_2_Chars( xfer0.n_files, g_slist_length(xfer->file_queue) );
-  DW_2_Chars( xfer0.n_bytes, xfer->total_bytes );
-  DW_2_Chars( xfer0.speed, 0x00000064 );
-  Word_2_Chars( xfer0.nick_len, strlen( our_info->nick ) + 1 );
-
-  fcntl( xfer->sock_fd, O_NONBLOCK );
+  /*   fcntl( xfer->sock_fd, O_NONBLOCK ); */
 
   xfer->gioc = g_io_channel_unix_new(xfer->sock_fd);
 
@@ -111,6 +119,7 @@ int ft_connectfile( XferInfo *xfer )
         (GIOFunc) ft_filereadserver,
         xfer, NULL );
 
+  /*
   index = 0;
   psize = sizeof( handshake );
   buffer = (BYTE*)g_malloc( psize + 2 );
@@ -123,19 +132,7 @@ int ft_connectfile( XferInfo *xfer )
                 PACKET_TYPE_TCP | PACKET_DIRECTION_SEND, "FILE(INIT)" );
   write( xfer->sock_fd, buffer, psize + 2 );
   g_free( buffer );
-
-  index = 0;
-  psize = sizeof( xfer0 ) + strlen( our_info->nick ) + 1;
-  buffer = (BYTE*)g_malloc( psize );
-  
-  memcpy( buffer, &xfer0, sizeof( xfer0 ) );
-  index += sizeof( xfer0 );
-  memcpy( buffer + index, our_info->nick, strlen( our_info->nick ) + 1 );
-
-  packet_print( buffer, psize,
-                PACKET_TYPE_TCP | PACKET_DIRECTION_SEND, "FILE(INIT)" );
-  write( xfer->sock_fd, buffer, psize );
-  g_free( buffer );
+  */
 
   return xfer->sock_fd;
 }
@@ -154,14 +151,14 @@ int ft_handshake( GIOChannel *source, GIOCondition condition,
   new_sock = accept( xfer->sock_fd, ( struct sockaddr * )&their_addr, &size );
 
         if (xfer->gioc)
-                g_io_channel_close(xfer->gioc);
+                g_io_channel_shutdown(xfer->gioc, TRUE, NULL);
 
   xfer->sock_fd = new_sock;
   xfer->gioc = g_io_channel_unix_new(new_sock);
 
   xfer->port = g_ntohs( their_addr.sin_port );
 
-  fcntl( new_sock, O_NONBLOCK );
+  /*fcntl( new_sock, O_NONBLOCK ); */
 
   g_io_add_watch_full( xfer->gioc, G_PRIORITY_DEFAULT_IDLE,
           G_IO_IN | G_IO_ERR | G_IO_HUP,
@@ -178,15 +175,29 @@ int ft_filereadserver(GIOChannel *source, GIOCondition condition,
   unsigned short packet_size;
   BYTE *packet;
 
+  struct {
+    BYTE psize[2];
+    BYTE id; /* 0x00 */
+    BYTE zero[4]; /* 0x00000000 */
+    BYTE n_files[4]; /* Number of files */
+    BYTE n_bytes[4]; /* Total number of bytes */
+    BYTE speed[4]; /* Speed of transfer, 0x00 - 0x64 */
+    BYTE nick_len[2]; /* Length of nick */
+    /* Nickname */
+  } xfer0;
+  BYTE *buffer;
+  int index, psize;
+
+
 #ifdef TRACE_FUNCTION
   g_print( "ft_filereadserver\n" );
 #endif
 
-       /* We closed the socket, so the transfer's already been cancelled  */
-       if (condition & G_IO_NVAL)
-               return FALSE;
-
-       if (read( xfer->sock_fd, (char*)(&packet_size), 2 ) != 2) {
+  /* We closed the socket, so the transfer's already been cancelled  */
+  if (condition & G_IO_NVAL)
+    return FALSE;
+  
+  if (read( xfer->sock_fd, (char*)(&packet_size), 2 ) != 2) {
     ft_cancel_transfer(0, xfer);
     return FALSE;
   }
@@ -198,38 +209,66 @@ int ft_filereadserver(GIOChannel *source, GIOCondition condition,
   packet_print( packet, packet_size,
                 PACKET_TYPE_TCP | PACKET_DIRECTION_RECEIVE, "FILE(INIT(READ))" );
 
-  switch( packet[ 0 ] )
-  {
+  switch( packet[ 0 ] ) {
+  case 0xFF: /* handshake */
+
+    TCPSendHandShakeAck(xfer->sock_fd);
+
+    Word_2_Chars( xfer0.psize, sizeof( xfer0 ) - 2 + strlen( our_info->nick ) + 1 );
+    xfer0.id = 0x00;
+    DW_2_Chars( xfer0.zero, 0x00000000 );
+    DW_2_Chars( xfer0.n_files, g_slist_length(xfer->file_queue) );
+    DW_2_Chars( xfer0.n_bytes, xfer->total_bytes );
+    DW_2_Chars( xfer0.speed, 0x00000064 );
+    Word_2_Chars( xfer0.nick_len, strlen( our_info->nick ) + 1 );
+
+    index = 0;
+    psize = sizeof( xfer0 ) + strlen( our_info->nick ) + 1;
+    buffer = (BYTE*)g_malloc( psize );
+    
+    memcpy( buffer, &xfer0, sizeof( xfer0 ) );
+    index += sizeof( xfer0 );
+    memcpy( buffer + index, our_info->nick, strlen( our_info->nick ) + 1 );
+    
+    packet_print( buffer, psize,
+		  PACKET_TYPE_TCP | PACKET_DIRECTION_SEND, "FILE(INIT)" );
+    write( xfer->sock_fd, buffer, psize );
+    g_free( buffer );
+
+    
+    break;
+
   case 0x01: /* We need to send XFER2 now */
-    ft_sendfileinfo( xfer );
+    if (packet_size != 4)
+      ft_sendfileinfo( xfer );
     break;
 
   case 0x03: /* Start sending the file :) */
-  /* The best way to do this is to set up an IO watch to send it so we don't
-     freeze the program, especially on large files */
-    ((struct file_struct*)xfer->file_queue->data)->completed_bytes = Chars_2_DW (&packet[1]);
+    /* The best way to do this is to set up an IO watch to send it so we don't
+       freeze the program, especially on large files */
+    ((struct xfer_file*)xfer->file_queue->data)->completed_bytes = Chars_2_DW (&packet[1]);
 
-    xfer->file_fd = open( ((struct file_struct*)xfer->file_queue->data)->full_filename, O_RDONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
+    xfer->file_fd = open( ((struct xfer_file*)xfer->file_queue->data)->full_filename, O_RDONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
     if( xfer->file_fd == -1 )
-    {
-      g_free(packet);
-      return ft_next_file(xfer);
-    }
+      {
+	g_free(packet);
+	return ft_next_file(xfer);
+      }
 
-    lseek( xfer->file_fd, ((struct file_struct*)xfer->file_queue->data)->completed_bytes, SEEK_SET );
+    lseek( xfer->file_fd, ((struct xfer_file*)xfer->file_queue->data)->completed_bytes, SEEK_SET );
 
     xfer->dialog = create_file_xfer_dialog(xfer);
     if (xfer->dialog == NULL) {
-       ft_cancel_transfer(0, xfer);
-       return FALSE;
+      ft_cancel_transfer(0, xfer);
+      return FALSE;
     }
 
     g_io_add_watch_full( xfer->gioc, G_PRIORITY_DEFAULT_IDLE,
-      G_IO_OUT | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-      (GIOFunc) ft_sendpartfile,
-      xfer, NULL );
+			 G_IO_OUT | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+			 (GIOFunc) ft_sendpartfile,
+			 xfer, NULL );
 
-    ((struct file_struct *)xfer->file_queue->data)->start_time = time( NULL );
+    ((struct xfer_file *)xfer->file_queue->data)->start_time = time( NULL );
 
     if( xfer->start_time == 0 )
       xfer->start_time = time( NULL );
@@ -240,6 +279,10 @@ int ft_filereadserver(GIOChannel *source, GIOCondition condition,
   case 0x04:
     ft_skip_file(0, xfer);
     break;
+  case 0x05:  /* looks like an ACK */
+    break;
+  default:
+    g_warning("Unknown packet for file transfer: 0x%02x\n", packet[ 0 ]);
   }
 
   g_free( packet );
@@ -274,7 +317,7 @@ int ft_filereadclient( GIOChannel *source, GIOCondition condition,
     /* We have to turn off blocking for just a bit, or else
     file transfers can get a bit flakey */
 
-    fcntl( xfer->sock_fd, F_SETFL, 0 );
+    /* fcntl( xfer->sock_fd, F_SETFL, 0 ); */
 
     if (read( xfer->sock_fd, (char*)(&xfer->packet_size), 2 ) != 2) {
       ft_cancel_transfer(0, xfer);
@@ -282,7 +325,7 @@ int ft_filereadclient( GIOChannel *source, GIOCondition condition,
     }
 
     /* Alright, don't block anymore */
-    fcntl( xfer->sock_fd, F_SETFL, 0 );
+    /* fcntl( xfer->sock_fd, F_SETFL, O_NONBLOCK ); */ 
 
     xfer->packet_size = GUINT16_FROM_LE (xfer->packet_size);
 
@@ -313,7 +356,7 @@ int ft_filereadclient( GIOChannel *source, GIOCondition condition,
       xfer->total_bytes = Chars_2_DW( &xfer->packet[ 9 ] );
 
       for(;file_count;file_count--)
-        xfer->file_queue = g_slist_append(xfer->file_queue, g_malloc0(sizeof(struct file_struct)));
+        xfer->file_queue = g_slist_append(xfer->file_queue, g_malloc0(sizeof(struct xfer_file)));
 
       xfer1.id = 0x01;
       DW_2_Chars( xfer1.speed, 0x00000064 );
@@ -333,8 +376,8 @@ int ft_filereadclient( GIOChannel *source, GIOCondition condition,
         if (!ft_next_file(xfer))
           return FALSE;
 
-      ((struct file_struct *)xfer->file_queue->data)->short_filename = g_strdup(&xfer->packet[ 4 ]);
-      ((struct file_struct *)xfer->file_queue->data)->total_bytes = Chars_2_DW(&xfer->packet[ 8 + strlen( &xfer->packet[4]) ]);
+      ((struct xfer_file *)xfer->file_queue->data)->short_filename = g_strdup(&xfer->packet[ 4 ]);
+      ((struct xfer_file *)xfer->file_queue->data)->total_bytes = Chars_2_DW(&xfer->packet[ 8 + strlen( &xfer->packet[4]) ]);
 
       file_place( xfer );     
 
@@ -348,7 +391,7 @@ int ft_filereadclient( GIOChannel *source, GIOCondition condition,
 
       write( xfer->file_fd, &xfer->packet[ 1 ], xfer->packet_size - 1);
 
-      ((struct file_struct *)xfer->file_queue->data)->completed_bytes += xfer->packet_size - 1;
+      ((struct xfer_file *)xfer->file_queue->data)->completed_bytes += xfer->packet_size - 1;
       xfer->completed_bytes += xfer->packet_size - 1;
 
     break;
@@ -364,10 +407,15 @@ int ft_filereadclient( GIOChannel *source, GIOCondition condition,
     if (contact == NULL)
       break;
     
-    TCPSendHandShakeAck(contact, xfer->sock_fd);
-    TCPSendHandShake(contact, Chars_2_DW(xfer->packet+32), xfer->sock_fd);
+    TCPSendHandShakeAck(xfer->sock_fd);
+    TCPSendHandShake(kontakt, Chars_2_DW(xfer->packet+32), xfer->sock_fd);
 
     break;
+  case 0x01:
+    if (xfer->packet_size == 4)
+      break;
+  default:
+    g_warning("Unknown transfer packet: 0x%02x\n", (guchar) xfer->packet[ 0 ]);
   }
 
   g_free(xfer->packet);
@@ -403,22 +451,22 @@ void ft_sendfileinfo( XferInfo *xfer )
 
   xfer2a.id = 0x02;
   xfer2a.zero = 0x00;
-  Word_2_Chars( xfer2a.fn_len, strlen( ((struct file_struct*)xfer->file_queue->data)->short_filename ) + 1 );
+  Word_2_Chars( xfer2a.fn_len, strlen( ((struct xfer_file*)xfer->file_queue->data)->short_filename ) + 1 );
   Word_2_Chars( xfer2b.one, 0x0001 );
   xfer2b.zero = 0;
-  DW_2_Chars( xfer2b.file_size, ((struct file_struct*)xfer->file_queue->data)->total_bytes );
+  DW_2_Chars( xfer2b.file_size, ((struct xfer_file*)xfer->file_queue->data)->total_bytes );
   DW_2_Chars( xfer2b.longzero, 0x00000000 );
   DW_2_Chars( xfer2b.speed, 0x00000064 );
     
-  packet_size = sizeof( xfer2a ) + sizeof( xfer2b ) + strlen( ((struct file_struct*)xfer->file_queue->data)->short_filename ) + 1;
+  packet_size = sizeof( xfer2a ) + sizeof( xfer2b ) + strlen( ((struct xfer_file*)xfer->file_queue->data)->short_filename ) + 1;
   index = 0;
   buffer = g_malloc( packet_size + 2 );
   Word_2_Chars (buffer, packet_size);
   index += 2;
   memcpy( buffer + index, &xfer2a, sizeof( xfer2a ) );
   index += sizeof( xfer2a );
-  memcpy( buffer + index, ((struct file_struct*)xfer->file_queue->data)->short_filename, strlen( ((struct file_struct*)xfer->file_queue->data)->short_filename ) + 1 );
-  index += strlen( ((struct file_struct*)xfer->file_queue->data)->short_filename ) + 1;
+  memcpy( buffer + index, ((struct xfer_file*)xfer->file_queue->data)->short_filename, strlen( ((struct xfer_file*)xfer->file_queue->data)->short_filename ) + 1 );
+  index += strlen( ((struct xfer_file*)xfer->file_queue->data)->short_filename ) + 1;
   memcpy( buffer + index, &xfer2b, sizeof( xfer2b ) );
   index += sizeof( xfer2b );
 
@@ -458,10 +506,10 @@ int ft_sendpartfile( GIOChannel *source, GIOCondition condition, XferInfo *xfer
   write( xfer->sock_fd, &xfer6, 3 + bytes );
   packet_print( &xfer6.id, bytes + 1, PACKET_TYPE_TCP | PACKET_DIRECTION_SEND, "FILE PACKET" );
       
-  ((struct file_struct *)xfer->file_queue->data)->completed_bytes += bytes;
+  ((struct xfer_file *)xfer->file_queue->data)->completed_bytes += bytes;
   xfer->completed_bytes += bytes;
 
-  if ((((struct file_struct *)xfer->file_queue->data)->completed_bytes >= ((struct file_struct *)xfer->file_queue->data)->total_bytes) || (bytes < 2048) ) {
+  if ((((struct xfer_file *)xfer->file_queue->data)->completed_bytes >= ((struct xfer_file *)xfer->file_queue->data)->total_bytes) || (bytes < 2048) ) {
     ft_next_file(xfer);
     return FALSE;
   }
@@ -482,7 +530,7 @@ void ft_cancel_transfer( GtkWidget *widget, XferInfo *xfer )
 #endif
 
   if( xfer->gioc )
-    g_io_channel_close(xfer->gioc);
+    g_io_channel_shutdown(xfer->gioc, FALSE, NULL);
 
   destroy_file_xfer_dialog(xfer->dialog);
 
@@ -490,7 +538,7 @@ void ft_cancel_transfer( GtkWidget *widget, XferInfo *xfer )
     if( xfer->direction == XFER_DIRECTION_RECEIVING ) {
       /* <kludge> */
       /*  Grrr, or else the last file won't be counted */
-      struct file_struct *completed_file = (struct file_struct *)xfer->file_queue->data;
+      struct xfer_file *completed_file = (struct xfer_file *)xfer->file_queue->data;
 
       xfer->file_queue = g_slist_remove( xfer->file_queue, completed_file );
       xfer->completed_files = g_slist_append( xfer->completed_files, completed_file );
@@ -514,7 +562,7 @@ void ft_cancel_transfer( GtkWidget *widget, XferInfo *xfer )
   }
   else{
     if (xfer->status == XFER_STATUS_PENDING) {
-      pending_xfers = g_slist_remove(pending_xfers, xfer);
+      xfer->remote_contact->file_xfers = g_slist_remove(xfer->remote_contact->file_xfers, xfer);
       msg = g_strdup_printf( _("File transfer refused: %s"), xfer->filename );
       gnome_ok_dialog( msg );
       g_free( msg );
@@ -538,18 +586,18 @@ void ft_cancel_transfer( GtkWidget *widget, XferInfo *xfer )
     close( xfer->file_fd );
 
   for(file_queue = xfer->file_queue;file_queue;file_queue = file_queue->next) {
-    g_free( ((struct file_struct*)file_queue->data)->short_filename);
-    g_free( ((struct file_struct*)file_queue->data)->full_filename);
-    g_free( (struct file_struct*)file_queue->data );
+    g_free( ((struct xfer_file*)file_queue->data)->short_filename);
+    g_free( ((struct xfer_file*)file_queue->data)->full_filename);
+    g_free( (struct xfer_file*)file_queue->data );
   }
 
   g_slist_free(xfer->file_queue);
 
   for(file_queue = xfer->completed_files;file_queue;
       file_queue = file_queue->next) {
-    g_free( ((struct file_struct*)file_queue->data)->short_filename);
-    g_free( ((struct file_struct*)file_queue->data)->full_filename);
-    g_free( (struct file_struct*)file_queue->data );
+    g_free( ((struct xfer_file*)file_queue->data)->short_filename);
+    g_free( ((struct xfer_file*)file_queue->data)->full_filename);
+    g_free( (struct xfer_file*)file_queue->data );
   }
 
   g_slist_free(xfer->completed_files);
@@ -584,7 +632,7 @@ void ft_skip_file( GtkWidget *widget, XferInfo *xfer)
 int ft_next_file( XferInfo *xfer )
 {
 
-  struct file_struct *completed_file = (struct file_struct *)xfer->file_queue->data;
+  struct xfer_file *completed_file = (struct xfer_file *)xfer->file_queue->data;
 
 #ifdef TRACE_FUNCTION
   g_print( "ft_next_file\n" );
@@ -634,14 +682,14 @@ void ft_readytoreceive (XferInfo *xfer)
 
   Word_2_Chars( xfer3.psize, 0x0d );
   xfer3.id = 0x03;
-  if( stat( ((struct file_struct*)xfer->file_queue->data)->full_filename, &file_stat ) == -1 )
-    ((struct file_struct *)xfer->file_queue->data)->completed_bytes = 0;
+  if( stat( ((struct xfer_file*)xfer->file_queue->data)->full_filename, &file_stat ) == -1 )
+    ((struct xfer_file *)xfer->file_queue->data)->completed_bytes = 0;
   else
-    ((struct file_struct *)xfer->file_queue->data)->completed_bytes = file_stat.st_size;
+    ((struct xfer_file *)xfer->file_queue->data)->completed_bytes = file_stat.st_size;
         
-  xfer->completed_bytes += ((struct file_struct *)xfer->file_queue->data)->completed_bytes;
+  xfer->completed_bytes += ((struct xfer_file *)xfer->file_queue->data)->completed_bytes;
         
-  DW_2_Chars( xfer3.resume_loc, ((struct file_struct *)xfer->file_queue->data)->completed_bytes );
+  DW_2_Chars( xfer3.resume_loc, ((struct xfer_file *)xfer->file_queue->data)->completed_bytes );
         
   DW_2_Chars( xfer3.longzero, 0x00000000 );
   DW_2_Chars( xfer3.speed, 0x00000064 );
@@ -658,7 +706,9 @@ gboolean ft_listen( XferInfo *xfer)
   g_print( "ft_listen\n" );
 #endif
 
-	pending_xfers = g_slist_remove(pending_xfers, xfer);
+	g_assert(xfer != NULL);
+
+	xfer->remote_contact->file_xfers = g_slist_remove(xfer->remote_contact->file_xfers, xfer);
 	xfer->status = XFER_STATUS_CONNECTING;
   
 	xfer->sock_fd = socket( PF_INET, SOCK_STREAM, 0 );
@@ -668,7 +718,7 @@ gboolean ft_listen( XferInfo *xfer)
 		return FALSE;
 	}
   
-	fcntl( xfer->sock_fd, O_NONBLOCK );
+	/* fcntl( xfer->sock_fd, O_NONBLOCK ); */
 
 	my_addr.sin_family = AF_INET;
 	my_addr.sin_port = 0;
@@ -697,3 +747,40 @@ gboolean ft_listen( XferInfo *xfer)
 
   return TRUE;
 }
+
+XferInfo* ft_new(Contact_Member *contact, int direction,
+		 char *filename, guint filesize) 
+{
+  XferInfo *xferinfo;
+  xferinfo = g_new0(XferInfo, 1);
+
+  xferinfo->remote_contact = contact;
+  
+  xferinfo->direction = direction;
+  xferinfo->status = XFER_STATUS_PENDING;
+
+  xferinfo->filename = filename;
+  xferinfo->total_bytes = filesize;
+
+  contact->file_xfers = g_slist_append(contact->file_xfers, xferinfo);
+
+  return xferinfo;
+}
+
+void ft_addfile(XferInfo *xferinfo, gchar *longname, gchar *shortname,
+		guint size)
+{
+
+  struct xfer_file *fileinfo;
+  
+  fileinfo = g_malloc0(sizeof(struct xfer_file));
+  
+  fileinfo->full_filename = longname;
+  fileinfo->short_filename = shortname;
+  
+  fileinfo->total_bytes = size;
+  xferinfo->total_bytes += size;
+  
+  xferinfo->file_queue = g_slist_append(xferinfo->file_queue, fileinfo);
+  
+}
